"""
This is the entry point for the `mailrise` command-line program.
"""

from __future__ import annotations

import argparse
import logging
import signal
import ssl
import sys
import typing as typ
from asyncio.events import new_event_loop
from functools import partial

from aiosmtpd.controller import UnthreadedController

from mailrise import __version__
from mailrise.config import TLSMode, load_config
from mailrise.smtp import AppriseHandler

__author__ = "Ryan Young"
__copyright__ = "Ryan Young"
__license__ = "MIT"

_logger = logging.getLogger(__name__)


# ---- Python API ----
# The functions defined in this section can be imported by users in their
# Python scripts/interactive interpreter, e.g. via
# `from mailrise.skeleton import fib`,
# when using this Python module as a library.


# ---- CLI ----
# The functions defined in this section are wrappers around the main Python
# API allowing them to be called directly from the terminal as a CLI
# executable/script.


def parse_args(args: list[str]) -> argparse.Namespace:
    """Parse command line parameters.

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace.
    """
    parser = argparse.ArgumentParser(
        description="An SMTP gateway for Apprise notifications")
    parser.add_argument(
        "--version",
        action="version",
        version=f"mailrise {__version__}"
    )
    parser.add_argument(
        dest="config",
        help="path to configuration file",
        type=argparse.FileType("r"),
        metavar="CONFIG"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel: int) -> None:
    """Setup basic logging.

    Args:
      loglevel (int): Minimum loglevel for emitting messages.
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args: list[str]) -> None:
    """Loads the configuration specified on the command-line and starts an SMTP
    server.

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--verbose", "42"]``).
    """
    pargs = parse_args(args)
    setup_logging(pargs.loglevel)

    config = load_config(_logger, pargs.config)

    tls: typ.Optional[ssl.SSLContext] = None
    tls_mode = config.tls_mode
    if tls_mode != TLSMode.OFF:
        assert isinstance(config.tls_certfile, str)
        assert isinstance(config.tls_keyfile, str)
        tls = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        tls.load_cert_chain(config.tls_certfile, keyfile=config.tls_keyfile)
        _logger.info('TLS enabled and successfully initialized')
    tls_onconnect = tls if tls_mode == TLSMode.ONCONNECT else None
    tls_starttls = \
        tls if tls_mode in (TLSMode.STARTTLS, TLSMode.STARTTLSREQUIRE) else None

    makecon = partial(
        UnthreadedController,
        AppriseHandler(config=config),
        authenticator=config.authenticator,
        auth_required=config.authenticator is not None,
        # We assume that if you've enabled STARTTLS, you'll want to require it.
        auth_require_tls=tls_starttls is not None,
        hostname=config.listen_host,
        port=config.listen_port,
        server_hostname=config.smtp_hostname,
        decode_data=False,
        ident=f'Mailrise {__version__}',
        tls_context=tls_starttls,
        ssl_context=tls_onconnect,
        require_starttls=tls_mode == TLSMode.STARTTLSREQUIRE
    )
    _logger.debug(
        'Arguments for aiosmtpd: %s',
        ', '.join(f'{kw}={makecon.keywords[kw]}'
                  for kw in ('authenticator', 'auth_required', 'auth_require_tls',
                             'tls_context', 'ssl_context', 'require_starttls')))

    eloop = new_event_loop()
    controller = makecon(loop=eloop)

    def clean_exit():
        _logger.info('Caught exit signal...')
        eloop.stop()
        controller.end()

    try:
        for sig in (signal.SIGINT, signal.SIGTERM):
            eloop.add_signal_handler(sig, clean_exit)
    except NotImplementedError:
        # add_signal_handler() is exclusive to Unix. No big deal if we can't set
        # up a handler.
        pass

    controller.begin()
    eloop.run_forever()


def run() -> None:
    """Calls :func:`main` passing the CLI arguments extracted from :obj:`sys.argv`.

    This function can be used as entry point to create console scripts with setuptools.
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    # ^  This is a guard statement that will prevent the following code from
    #    being executed in the case someone imports this file instead of
    #    executing it as a script.
    #    https://docs.python.org/3/library/__main__.html

    # After installing your project with pip, users can also run your Python
    # modules as scripts via the ``-m`` flag, as defined in PEP 338::
    #
    #     python -m mailrise.skeleton 42
    #
    run()



"""
This is the SMTP server functionality for Mailrise.
"""

from __future__ import annotations

import asyncio
import email.policy
import os
import typing as typ
from email import contentmanager
from email.message import EmailMessage as StdlibEmailMessage
from email.parser import BytesParser
from tempfile import NamedTemporaryFile

import apprise
from aiosmtpd.smtp import Envelope, Session, SMTP
from apprise.common import ContentLocation

from mailrise.config import MailriseConfig
import mailrise.router as r

# Mypy, for some reason, considers AttachBase a module, not a class.
MYPY = False
# pylint: disable=ungrouped-imports
if MYPY:
    from apprise.attachment.AttachBase import AttachBase
else:
    from apprise.attachment import AttachBase


class AppriseNotifyFailure(Exception):
    """Exception raised when Apprise fails to deliver a notification.

    Note: Apprise does not provide any information about the reason for the
    failure.
    """


class UnreadableMultipart(Exception):
    """Exception raised for multipart messages that can't be parsed.

    Attributes:
        message: The multipart email part.
    """
    message: StdlibEmailMessage

    def __init__(self, message: StdlibEmailMessage) -> None:
        super().__init__(self)
        self.message = message


class AppriseHandler(typ.NamedTuple):
    """The aiosmtpd handler for Mailrise. Dispatches Apprise notifications.

    Attributes:
        config: This server's Mailrise configuration.
    """
    config: MailriseConfig

    # pylint: disable=invalid-name,unused-argument,too-many-arguments
    async def handle_RCPT(self, server: SMTP, session: Session, envelope: Envelope,
                          address: str, rcpt_options: list[str]) -> str:
        """Called during RCPT TO."""
        self.config.logger.info('Added recipient: %s', address)
        envelope.rcpt_tos.append(address)
        return '250 OK'

    # pylint: disable=invalid-name,unused-argument
    async def handle_DATA(self, server: SMTP, session: Session, envelope: Envelope) \
            -> str:
        """Called during DATA after the entire message ('SMTP content' as described
        in RFC 5321) has been received."""
        assert isinstance(envelope.content, bytes)
        parser = BytesParser(policy=email.policy.default)
        message = parser.parsebytes(envelope.content)
        assert isinstance(message, StdlibEmailMessage)
        try:
            notification = _parsemessage(message, envelope)
        except UnreadableMultipart as mpe:
            subparts = \
                ' '.join(part.get_content_type() for part in mpe.message.iter_parts())
            self.config.logger.error('Failed to parse %s message: [ %s ]',
                                     mpe.message.get_content_type(), subparts)
        self.config.logger.info('Accepted email, subject: %s', notification.subject)

        try:
            to_send = [data async for data in self.config.router.email_to_apprise(
                           logger=self.config.logger,
                           email=notification,
                           auth_data=session.auth_data
                       )]
        except Exception as exc:  # pylint: disable=broad-except
            return f'450 router had internal exception: {exc}'

        results = await asyncio.gather(
            *(_apprise_notify(self.config, data) for data in to_send),
            return_exceptions=True
        )
        if any(isinstance(result, AppriseNotifyFailure) for result in results):
            addresses = ' '.join(envelope.rcpt_tos)
            self.config.logger.warning('Notification failed: %s ➤ %s',
                                       notification.subject, addresses)
            return '450 failed to send notification'

        return '250 OK'


def _parsemessage(msg: StdlibEmailMessage, envelope: Envelope) -> r.EmailMessage:
    """Parses an email message into an `EmailNotification`.

    Args:
        msg: The email message.

    Returns:
        The `EmailNotification` instance.
    """
    py_body_part = msg.get_body()
    body: typ.Optional[tuple[str, apprise.NotifyFormat]]
    if isinstance(py_body_part, StdlibEmailMessage):
        body_part: StdlibEmailMessage
        try:
            py_body_part.get_content()
        except KeyError:  # stdlib failed to read the content, which means multipart
            body_part = _getmultiparttext(py_body_part)
        else:
            body_part = py_body_part
        body_content = contentmanager.raw_data_manager.get_content(body_part)
        is_html = body_part.get_content_subtype() == 'html'
        body = (body_content.strip(),
                apprise.NotifyFormat.HTML if is_html else apprise.NotifyFormat.TEXT)
    else:
        body = None
    attachments = [_parseattachment(part) for part in msg.iter_attachments()
                   if isinstance(part, StdlibEmailMessage)]
    return r.EmailMessage(
        email_message=msg,
        subject=msg.get('Subject', '[no subject]'),
        from_=msg.get('From', '[no sender]'),
        to=envelope.rcpt_tos,
        # Apprise will fail if no body is supplied.
        body=body[0] if body else '[no body]',
        body_format=body[1] if body else apprise.NotifyFormat.TEXT,
        attachments=attachments
    )


def _getmultiparttext(msg: StdlibEmailMessage) -> StdlibEmailMessage:
    """Search for the textual body part of a multipart email."""
    content_type = msg.get_content_type()
    if content_type in ('multipart/related', 'multipart/alternative'):
        parts = list(msg.iter_parts())
        # Look for these types of parts in descending order.
        for parttype in ('multipart/alternative', 'multipart/related',
                         'text/html', 'text/plain'):
            found = \
                next((p for p in parts if isinstance(p, StdlibEmailMessage)
                     and p.get_content_type() == parttype), None)
            if found is not None:
                return _getmultiparttext(found)
        raise UnreadableMultipart(msg)
    return msg


def _parseattachment(part: StdlibEmailMessage) -> r.EmailAttachment:
    return r.EmailAttachment(data=part.get_content(), filename=part.get_filename(''))


async def _apprise_notify(config: MailriseConfig, data: r.AppriseNotification):
    ap_config = apprise.AppriseConfig(asset=data.asset or r.DEFAULT_ASSET)
    ap_config.add_config(data.config, format=data.config_format)
    ap_instance = apprise.Apprise(ap_config)

    attach_base = [_AttachMailrise(config, attach) for attach in data.attachments]
    success = await ap_instance.async_notify(
        title=data.title,
        body=data.body,
        body_format=data.body_format,
        notify_type=data.notify_type,
        attach=attach_base
    )
    # NOTE: This should probably be called by Apprise itself, but it isn't?
    for base in attach_base:
        base.invalidate()
    if not success:
        raise AppriseNotifyFailure


class _AttachMailrise(AttachBase):
    """An Apprise attachment type that wraps `Attachment`.

    Data is stored in temporary files for upload.

    Args:
        config: The Mailrise configuration to use.
        attach: The `Attachment` instance.
    """
    location = ContentLocation.LOCAL

    _mrfile = None  # Satisfy mypy by initializing as an Optional.

    def __init__(self, config: MailriseConfig,
                 attach: r.EmailAttachment, **kwargs: typ.Any) -> None:
        super().__init__(**kwargs)
        self._mrconfig = config
        self._mrattach = attach

    def download(self) -> bool:
        self.invalidate()

        with NamedTemporaryFile(delete=False) as tfile:
            tfile.write(self._mrattach.data)
        self._mrfile = tfile
        self.download_path = tfile.name
        self.detected_name = self._mrattach.filename

        return True  # Indicates the "download" was successful.

    def invalidate(self) -> None:
        tfile = self._mrfile
        if tfile:
            try:
                os.remove(tfile.name)
            except (FileNotFoundError, OSError):
                self._mrconfig.logger.info(
                    'Failed to delete attachment file: %s', tfile.name)
            self._mrfile = None
        super().invalidate()

    def url(self, *args: typ.Any, **kwargs: typ.Any) -> str:
        return f'mailrise://{hex(id(self))}'

    @staticmethod
    def parse_url(url: str, verify_host: bool = True,
                  mimetype_db: typ.Any = None) -> typ.Dict[str, typ.Any]:
        return {}



"""
Routers process email messages and produce notifications.
"""

import os
import typing as typ
from abc import ABCMeta, abstractmethod
from email.message import EmailMessage as StdlibEmailMessage
from logging import Logger

from apprise import AppriseAsset, NotifyFormat
from apprise.common import NotifyType


DEFAULT_ASSET = AppriseAsset(
    app_id='Mailrise',
    app_desc='Mailrise SMTP Notification Relay',
    app_url='https://mailrise.xyz',
    html_notify_map={
        NotifyType.INFO: '#2e6e99',
        NotifyType.SUCCESS: '#2e992e',
        NotifyType.WARNING: '#99972e',
        NotifyType.FAILURE: '#993a2e'
    },
    theme=None,
    default_extension='.png',
    image_url_mask='https://raw.githubusercontent.com/YoRyan/mailrise/main/'
                   'src/mailrise/asset/mailrise-{TYPE}-{XY}{EXTENSION}',
    image_url_logo='https://raw.githubusercontent.com/YoRyan/mailrise/main/'
                   'src/mailrise/asset/mailrise-logo.png',
    image_path_mask=os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            'asset',
            'mailrise-{TYPE}-{XY}{EXTENSION}'
        )
    )
)


class EmailAttachment(typ.NamedTuple):
    """Represents an email attachment.

    Attributes:
        data: The contents of the attachment.
        filename: The filename of the attachment as it was set by the sender.
    """
    data: bytes
    filename: str


class EmailMessage(typ.NamedTuple):
    """Represents an email accepted for notifying.

    Attributes:
        email_message: The raw, unprocessed email message as represented by the
            Python standard library.
        subject: The email subject.
        from_: The email sender address.
        to: The list of addresses the email is addressed to.
        body: The contents of the email.
        body_format: The type of the contents of the email.
        attachments: The email attachments.
    """
    email_message: StdlibEmailMessage
    subject: str
    from_: str
    to: typ.List[str]
    body: str
    body_format: NotifyFormat
    attachments: typ.List[EmailAttachment]


class AppriseNotification(typ.NamedTuple):
    """Encapsulates the information needed to submit a notification to Apprise.

    Attributes:
        config: The Apprise configuration file to use, as a string.
        title: The notification title.
        body: The notification body.
        notify_type: The class of notification (info/success/warning/failure).
        body_format: The content type (text/html/markdown) of the body.
        attachments: A list of attachments to send with the notification.
        config_format: The format of the configuration file. If None, Apprise
            will autodetect (with some slight overhead).
        asset: The Apprise asset, which controls the icon sent with the
            notification. If None, use some flavor of the Mailrise logo.
    """
    config: str
    title: str
    body: str
    notify_type: NotifyType = NotifyType.INFO
    body_format: NotifyFormat = NotifyFormat.TEXT
    attachments: typ.List[EmailAttachment] = []
    config_format: typ.Literal['text', 'yaml'] | None = None
    asset: AppriseAsset | None = None


class Router(metaclass=ABCMeta):  # pylint: disable=too-few-public-methods
    """A pluggable module that dispatches emails."""

    @abstractmethod
    async def email_to_apprise(
        self, logger: Logger, email: EmailMessage, auth_data: typ.Any, **kwargs) \
            -> typ.AsyncGenerator[AppriseNotification, None]:
        """Converts an email into one or multiple Apprise notifications."""
        # Needed to pass mypy, which fails to realize this is an async generator.
        # See https://github.com/python/mypy/issues/5070
        yield AppriseNotification(config='', title='', body='')



"""
This is the authentication functionality for the SMTP server.
"""

import typing as typ

from aiosmtpd import smtp


class BasicAuthenticator(typ.NamedTuple):
    """A simple authenticator that uses a static username and password list."""
    logins: typ.Mapping[str, str]

    # pylint: disable=too-many-arguments
    def __call__(self, server: smtp.SMTP, session: smtp.Session,
                 envelope: smtp.Envelope, mechanism: str, auth_data: typ.Any):
        fail_nothandled = smtp.AuthResult(success=False, handled=False)
        if mechanism not in ("LOGIN", "PLAIN"):
            return fail_nothandled
        if not isinstance(auth_data, smtp.LoginPassword):
            return fail_nothandled

        username = auth_data.login.decode("utf-8")
        password = auth_data.password.decode("utf-8")
        success = self.logins.get(username) == password
        return smtp.AuthResult(success=success)

    def __str__(self) -> str:
        return f'Basic({len(self.logins)})'



"""
This is the YAML-based router for Mailrise.
"""

from email.utils import parseaddr
from fnmatch import fnmatchcase
from logging import Logger
from string import Template
import re
import typing as typ

import apprise
import yaml

from mailrise.router import AppriseNotification, EmailMessage, Router


class _Key(typ.NamedTuple):
    """A unique identifier for a sender target.

    Attributes:
        user: The user portion of the recipient address.
        domain: The domain portion of the recipient address, which defaults
            to "mailrise.xyz".
    """
    user: str
    domain: str = 'mailrise.xyz'

    def __str__(self) -> str:
        return f'{self.user}@{self.domain}'

    def as_configured(self) -> str:
        """Drop the domain part of this identifier if it is 'mailrise.xyz'."""
        return self.user if self.domain == 'mailrise.xyz' else str(self)


class _Recipient(typ.NamedTuple):
    """The routing information encoded into a recipient address.

    Attributes:
        key: An index into the dictionary of senders.
        notify_type: The type of notification to send.
    """
    key: _Key
    notify_type: apprise.NotifyType


def _parsercpt(addr: str) -> _Recipient:
    _, rcpt = parseaddr(addr)
    user, domain = _parseaddrparts(rcpt)
    if not user or not domain:
        raise ValueError
    match = re.search(
        r'(.*)\.(info|success|warning|failure)$', user, re.IGNORECASE)
    ntype = apprise.NotifyType.INFO
    if match is not None:
        user = match.group(1)
        ntypes = match.group(2)
        if ntypes == 'info':
            pass
        elif ntypes == 'success':
            ntype = apprise.NotifyType.SUCCESS
        elif ntypes == 'warning':
            ntype = apprise.NotifyType.WARNING
        elif ntypes == 'failure':
            ntype = apprise.NotifyType.FAILURE
    return _Recipient(key=_Key(user=user, domain=domain.lower()), notify_type=ntype)


def _parseaddrparts(email: str) -> typ.Tuple[str, str]:
    """Parses an email address into its component user and domain parts."""
    match = re.search(r'(?:"([^"@]*)"|([^@]*))@([^@]*)$', email)
    if match is None:
        return '', ''
    quoted = match.group(1) is not None
    user = match.group(1) if quoted else match.group(2)
    domain = match.group(3)
    return user, domain


class _SimpleSender(typ.NamedTuple):
    """A configured target for Apprise notifications.

    Attributes:
        config_yaml: The YAML configuration for Apprise.
        title_template: The template string for notification title texts.
        body_template: The template string for notification body texts.
        body_format: The content type for notifications. If None, this will be
            auto-detected from the body parts of emails.
    """
    config_yaml: str
    title_template: Template
    body_template: Template
    body_format: typ.Optional[apprise.NotifyFormat]


class SimpleRouter(Router):  # pylint: disable=too-few-public-methods
    """A router that uses the rules in the YAML configuration file.

    Attributes:
        senders: A list of notification targets, each with a [key, sender]
            tuple, where key contains username and domain patterns that can be
            matched by fnmatch and sender is the Sender instance itself.
    """
    senders: typ.List[typ.Tuple[_Key, _SimpleSender]]

    def __init__(self, senders: typ.List[typ.Tuple[_Key, _SimpleSender]]):
        super().__init__()
        self.senders = senders

    async def email_to_apprise(
        self, logger: Logger, email: EmailMessage, auth_data: typ.Any, **kwargs) \
            -> typ.AsyncGenerator[AppriseNotification, None]:
        for addr in email.to:
            try:
                rcpt = _parsercpt(addr)
            except ValueError:
                logger.error('Not a valid Mailrise address: %s', addr)
                continue
            sender = self.get_sender(rcpt.key)
            if sender is None:
                logger.error('Recipient is not configured: %s', addr)
                continue

            mapping = {
                'subject': email.subject,
                'from': email.from_,
                'body': email.body,
                'to': str(rcpt.key),
                'config': rcpt.key.as_configured(),
                'type': rcpt.notify_type
            }
            yield AppriseNotification(
                config=sender.config_yaml,
                config_format='yaml',
                title=sender.title_template.safe_substitute(mapping),
                body=sender.body_template.safe_substitute(mapping),
                # Use the configuration body format if specified.
                body_format=sender.body_format or email.body_format,
                notify_type=rcpt.notify_type,
                attachments=email.attachments
            )

    def get_sender(self, key: _Key) -> _SimpleSender | None:
        """Find a sender by recipient key."""
        return next(
            (sender for (pattern_key, sender) in self.senders
             if fnmatchcase(key.user, pattern_key.user)
             and fnmatchcase(key.domain, pattern_key.domain)), None)


def load_from_yaml(logger: Logger, configs_node: dict[str, typ.Any]) -> SimpleRouter:
    """Load a simple router from the YAML configs node."""
    if not isinstance(configs_node, dict):
        logger.critical('The configs node is not a YAML mapping')
        raise SystemExit(1)
    router = SimpleRouter(
        senders=[(_parse_simple_key(logger, key), _load_simple_sender(logger, key, config))
                 for key, config in configs_node.items()]
    )
    if len(router.senders) < 1:
        logger.critical('No Apprise targets are configured')
        raise SystemExit(1)
    logger.info('Loaded configuration with %d recipient(s)', len(router.senders))
    return router


def _parse_simple_key(logger: Logger, key: str) -> _Key:
    def fatal():
        logger.critical(
            "Invalid config key '%s'; should be a string or an email address "
            'without periods in the username')
        raise SystemExit(1)
    if '@' in key:
        user, domain = _parseaddrparts(key)
        if not user or not domain or '.' in user:
            fatal()
        return _Key(user=user, domain=domain.lower())
    if '.' in key:
        fatal()

    return _Key(user=key)


def _load_simple_sender(logger: Logger, key: str, config: dict[str, typ.Any]) -> _SimpleSender:
    if not isinstance(config, dict):
        logger.critical("YAML config node '%s' is not a mapping", key)
        raise SystemExit(1)

    # Extract Mailrise-specific values.
    mr_config = config.get('mailrise', {})
    config.pop('mailrise', None)
    title_template = mr_config.get('title_template', '$subject ($from)')
    body_template = mr_config.get('body_template', '$body')
    body_format = mr_config.get('body_format', None)
    if not any(body_format == c for c in (None,
                                          apprise.NotifyFormat.TEXT,
                                          apprise.NotifyFormat.HTML,
                                          apprise.NotifyFormat.MARKDOWN)):
        logger.critical('Invalid Apprise notification format: %s', body_format)
        raise SystemExit(1)

    return _SimpleSender(
        config_yaml=yaml.safe_dump(config),
        title_template=Template(title_template),
        body_template=Template(body_template),
        body_format=body_format
    )



"""
Tests for configuration loading.
"""

import logging
from io import StringIO

import pytest
from apprise import Apprise, AppriseConfig, NotifyFormat
from pytest import MonkeyPatch

from mailrise.basic_authenticator import BasicAuthenticator
from mailrise.config import load_config
from mailrise.simple_router import _Key, SimpleRouter


_logger = logging.getLogger(__name__)


def test_errors() -> None:
    """Tests for :fun:`load_config`'s failure conditions."""
    with pytest.raises(SystemExit):
        file = StringIO("""
            24
        """)
        load_config(_logger, file)
    with pytest.raises(SystemExit):
        file = StringIO("""
            configs: 24
        """)
        load_config(_logger, file)
    with pytest.raises(SystemExit):
        file = StringIO("""
            configs:
              test: 24
        """)
        load_config(_logger, file)


def test_load() -> None:
    """Tests a successful load with :fun:`load_config`."""
    file = StringIO("""
        configs:
          test:
            urls:
              - json://localhost
    """)
    mrise = load_config(_logger, file)
    router = mrise.router
    assert isinstance(router, SimpleRouter)
    assert len(router.senders) == 1
    key = _Key(user='test')
    assert mrise.authenticator is None

    sender = router.get_sender(key)
    assert sender is not None
    notifier = _make_notifier(sender.config_yaml)
    assert len(notifier) == 1
    assert notifier[0].url().startswith('json://localhost/')


def test_multi_load() -> None:
    """Tests a sucessful load with :fun:`load_config` with multiple configs."""
    file = StringIO("""
        configs:
          test1:
            urls:
              - json://localhost
          test2:
            urls:
              - json://localhost
    """)
    mrise = load_config(_logger, file)
    router = mrise.router
    assert isinstance(router, SimpleRouter)
    assert len(router.senders) == 2

    for user in ('test1', 'test2'):
        key = _Key(user=user)

        sender = router.get_sender(key)
        assert sender is not None
        notifier = _make_notifier(sender.config_yaml)
        assert len(notifier) == 1
        assert notifier[0].url().startswith('json://localhost/')


def test_mailrise_options() -> None:
    """Tests a successful load with :fun:`load_config` with Mailrise-specific
    options."""
    file = StringIO("""
        configs:
          test:
            urls:
              - json://localhost
            mailrise:
              title_template: ""
              body_format: "text"
    """)
    mrise = load_config(_logger, file)
    router = mrise.router
    assert isinstance(router, SimpleRouter)
    assert len(router.senders) == 1
    key = _Key(user='test')

    sender = router.get_sender(key)
    assert sender is not None
    assert sender.title_template.template == ''
    assert sender.body_format == NotifyFormat.TEXT

    with pytest.raises(SystemExit):
        file = StringIO("""
            configs:
              test:
                urls:
                  - json://localhost
                mailrise:
                  body_format: "BAD"
        """)
        load_config(_logger, file)


def test_config_keys() -> None:
    """Tests the config key parser with both string and full email formats."""
    with pytest.raises(SystemExit):
        file = StringIO("""
            configs:
              has.periods:
                urls:
                  - json://localhost
        """)
        load_config(_logger, file)
    with pytest.raises(SystemExit):
        file = StringIO("""
            configs:
              bademail@:
                urls:
                  - json://localhost
        """)
        load_config(_logger, file)
    file = StringIO("""
        configs:
          user@example.com:
            urls:
              - json://localhost
    """)
    mrise = load_config(_logger, file)
    router = mrise.router
    assert isinstance(router, SimpleRouter)
    assert len(router.senders) == 1
    key = _Key(user='user', domain='example.com')
    assert router.get_sender(key) is not None


def test_fnmatch_config_keys() -> None:
    """Tests the config key parser with fnmatch pattern tokens."""
    # This defaults to "*@mailrise.xyz", which may not be obvious at first
    # glance.
    file = StringIO("""
        configs:
          "*":
            urls:
              - json://localhost
    """)
    mrise = load_config(_logger, file)
    router = mrise.router
    assert isinstance(router, SimpleRouter)
    key = _Key(user='user', domain='example.com')
    assert router.get_sender(key) is None
    key = _Key(user='user', domain='mailrise.xyz')
    assert router.get_sender(key) is not None

    file = StringIO("""
        configs:
          "*@*":
            urls:
              - json://localhost
    """)
    mrise = load_config(_logger, file)
    router = mrise.router
    assert isinstance(router, SimpleRouter)
    key = _Key(user='user', domain='example.com')
    assert router.get_sender(key) is not None

    file = StringIO("""
        configs:
          "the*@*":
            urls:
              - json://localhost
    """)
    mrise = load_config(_logger, file)
    router = mrise.router
    assert isinstance(router, SimpleRouter)
    key = _Key(user='user', domain='example.com')
    assert router.get_sender(key) is None
    key = _Key(user='thequickbrownfox', domain='example.com')
    assert router.get_sender(key) is not None


def test_authenticator() -> None:
    """Tests a successful load with an authenticator."""
    file = StringIO("""
        configs:
          test:
            urls:
              - json://localhost
        smtp:
          auth:
            basic:
              username: password
              AzureDiamond: hunter2
    """)
    mrise = load_config(_logger, file)
    assert isinstance(mrise.authenticator, BasicAuthenticator)
    logins = mrise.authenticator.logins
    assert logins['username'] == 'password'
    assert logins['AzureDiamond'] == 'hunter2'
    assert 'test' not in logins


def test_env_var() -> None:
    """Tests the environment variable loader."""
    with MonkeyPatch.context() as ctx:
        ctx.setenv('mytesturl', 'json://localhost')

        files = [
            StringIO("""
              configs:
                test:
                  urls:
                    - !env_var mytesturl
            """),
            StringIO("""
              configs:
                test:
                  urls:
                    - !env_var fallback json://localhost
            """)
        ]
        for file in files:
            mrise = load_config(_logger, file)
            router = mrise.router
            assert isinstance(router, SimpleRouter)
            assert len(router.senders) == 1
            key = _Key(user='test')
            sender = router.get_sender(key)
            assert sender is not None
            notifier = _make_notifier(sender.config_yaml)
            # Missing type annotation for this property as of Dec 2022.
            ap_servers = notifier.servers  # type: ignore
            assert len(ap_servers) == 1
            config = ap_servers[0]
            servers = config.servers()
            assert len(servers) == 1
            assert servers[0].url().startswith('json://localhost')

    with pytest.raises(SystemExit):
        file = StringIO("""
          configs:
            test:
              urls:
                - !env_var error
        """)
        load_config(_logger, file)


def _make_notifier(config: str):
    ap_config = AppriseConfig()
    ap_config.add_config(config, format='yaml')
    return Apprise(ap_config)



"""
Tests for the YAML-based router.
"""

import apprise
import pytest

from mailrise.simple_router import _Key, _parsercpt


def test_parsercpt() -> None:
    """Tests for recipient parsing."""
    rcpt = _parsercpt('test@mailrise.xyz')
    assert rcpt.key == _Key(user='test')
    assert rcpt.notify_type == apprise.NotifyType.INFO

    rcpt = _parsercpt('test.warning@mailrise.xyz')
    assert rcpt.key == _Key(user='test')
    assert rcpt.notify_type == apprise.NotifyType.WARNING

    rcpt = _parsercpt('"with_quotes"@mailrise.xyz')
    assert rcpt.key == _Key(user='with_quotes')
    assert rcpt.notify_type == apprise.NotifyType.INFO

    rcpt = _parsercpt('"with_quotes.success"@mailrise.xyz')
    assert rcpt.key == _Key('with_quotes')
    assert rcpt.notify_type == apprise.NotifyType.SUCCESS

    rcpt = _parsercpt('"weird_quotes".success@mailrise.xyz')
    assert rcpt.key == _Key('"weird_quotes"')
    assert rcpt.notify_type == apprise.NotifyType.SUCCESS

    rcpt = _parsercpt('John Doe <johndoe.warning@mailrise.xyz>')
    assert rcpt.key == _Key('johndoe')
    assert rcpt.notify_type == apprise.NotifyType.WARNING

    with pytest.raises(ValueError):
        _parsercpt("Invalid Email <bad@>")



"""
Tests for the SMTP server functionality.
"""

from email.message import EmailMessage
from pathlib import Path

import apprise
from aiosmtpd.smtp import Envelope

from mailrise.smtp import _parsemessage


def test_parsemessage() -> None:
    """Tests for email message parsing."""
    msg = EmailMessage()
    msg.set_content('Hello, World!')
    msg['From'] = ''
    msg['Subject'] = 'Test Message'
    notification = _parsemessage(msg, Envelope())
    assert notification.subject == 'Test Message'
    assert notification.body == 'Hello, World!'
    assert notification.body_format == apprise.NotifyFormat.TEXT

    msg = EmailMessage()
    msg.set_content('Hello, World!')
    msg.add_alternative('Hello, <strong>World!</strong>', subtype='html')
    notification = _parsemessage(msg, Envelope())
    assert notification.subject == '[no subject]'
    assert notification.from_ == '[no sender]'
    assert notification.body == 'Hello, <strong>World!</strong>'
    assert notification.body_format == apprise.NotifyFormat.HTML


def test_multipart() -> None:
    """Tests for email message parsing with multipart components."""
    img_name = 'bridge.jpg'
    with open(Path(__file__).parent/img_name, 'rb') as file:
        img_data = file.read()
    msg = EmailMessage()
    msg.add_related('Hello, World!')
    msg.add_related(img_data, maintype='image', subtype='jpeg')
    msg['From'] = ''
    msg['Subject'] = 'Test Message'
    notification = _parsemessage(msg, Envelope())
    assert notification.subject == 'Test Message'
    assert notification.body == 'Hello, World!'
    assert notification.body_format == apprise.NotifyFormat.TEXT

    msg = EmailMessage()
    msg.add_alternative('Hello, World!', subtype='plain')
    msg.add_alternative('<strong>Hello, World!</strong>', subtype='html')
    msg['From'] = ''
    msg['Subject'] = 'Test Message'
    notification = _parsemessage(msg, Envelope())
    assert notification.subject == 'Test Message'
    assert notification.body == '<strong>Hello, World!</strong>'
    assert notification.body_format == apprise.NotifyFormat.HTML


def test_parseattachments() -> None:
    """Tests for email message parsing with attachments."""
    img_name = 'bridge.jpg'
    with open(Path(__file__).parent/img_name, 'rb') as file:
        img_data = file.read()

    msg = EmailMessage()
    msg.set_content('Hello, World!')
    msg['From'] = 'sender@example.com'
    msg['Subject'] = 'Now With Images'
    msg.add_attachment(
        img_data,
        maintype='image',
        subtype='jpeg',
        filename=img_name
    )
    notification = _parsemessage(msg, Envelope())
    assert notification.subject == 'Now With Images'
    assert notification.from_ == 'sender@example.com'
    assert notification.body == 'Hello, World!'
    assert notification.body_format == apprise.NotifyFormat.TEXT
    assert len(notification.attachments) == 1
    assert notification.attachments[0].data == img_data
    assert notification.attachments[0].filename == img_name

    msg = EmailMessage()
    msg.set_content('Hello, World!')
    msg['From'] = 'sender@example.com'
    msg['Subject'] = 'Now With Images'
    msg.add_attachment(
        img_data,
        maintype='image',
        subtype='jpeg',
        filename=f'1_{img_name}'
    )
    msg.add_attachment(
        img_data,
        maintype='image',
        subtype='jpeg',
        filename=f'2_{img_name}'
    )
    notification = _parsemessage(msg, Envelope())
    assert notification.subject == 'Now With Images'
    assert notification.from_ == 'sender@example.com'
    assert notification.body == 'Hello, World!'
    assert notification.body_format == apprise.NotifyFormat.TEXT
    assert len(notification.attachments) == 2
    for attach in notification.attachments:
        assert attach.data == img_data
    assert notification.attachments[0].filename == f'1_{img_name}'
    assert notification.attachments[1].filename == f'2_{img_name}'



"""
This is the YAML configuration parser for Mailrise.
"""

from __future__ import annotations

import importlib.util
import io
import os
import typing as typ
from enum import Enum
from functools import partial
from logging import Logger
from typing import NamedTuple

import yaml
from aiosmtpd.smtp import AuthenticatorType

from mailrise.basic_authenticator import BasicAuthenticator
from mailrise.router import Router
from mailrise.simple_router import load_from_yaml as load_simple_router


class ConfigFileLoader(yaml.FullLoader):  # pylint: disable=too-many-ancestors
    """Our YAML loader class, which comes with an attached logger."""
    logger: Logger

    def __init__(self, stream, logger: Logger) -> None:
        super().__init__(stream)
        self.logger = logger
        self.add_constructor('!env_var', ConfigFileLoader._env_var_constructor)

    @staticmethod
    def _env_var_constructor(loader: ConfigFileLoader, node: yaml.nodes.Node) -> str:
        """Load environment variables and embed them into the configuration YAML."""
        value = str(node.value)
        try:
            env, default = value.split(maxsplit=1)
        except ValueError:
            env, default = value, None

        if env in os.environ:
            return os.environ[env]
        if default:
            loader.logger.warning(
                'Environment variable %s not defined, using default value: %s',
                env, default)
            return default
        loader.logger.critical(
            'Environment variable %s not defined and no default value provided', env)
        raise SystemExit(1)


class TLSMode(Enum):
    """Specifies a TLS encryption operating mode."""
    OFF = 'no TLS'
    ONCONNECT = 'TLS on connect'
    STARTTLS = 'STARTTLS, optional'
    STARTTLSREQUIRE = 'STARTTLS, required'


class MailriseConfig(NamedTuple):
    """Configuration data for a Mailrise instance.

    Attributes:
        logger: The logger, which is used to record interesting events.
        listen_host: The network address to listen on.
        listen_port: The network port to listen on.
        tls_mode: The TLS encryption mode.
        tls_certfile: The path to the TLS certificate chain file.
        tls_keyfile: The path to the TLS key file.
        smtp_hostname: The advertised SMTP server hostname.
        senders: A list of notification targets, each with a [key, sender]
            tuple, where key contains username and domain patterns that can be
            matched by fnmatch and sender is the Sender instance itself.
    """
    logger: Logger
    listen_host: str
    listen_port: int
    tls_mode: TLSMode
    tls_certfile: typ.Optional[str]
    tls_keyfile: typ.Optional[str]
    smtp_hostname: typ.Optional[str]
    router: Router
    authenticator: typ.Optional[AuthenticatorType]


class MailriseImportedCode(NamedTuple):
    """The pluggable Python code for a Mailrise instance when imported as a
    Python module. Of course, the actual result will be a module rather than
    a named tuple, but we can expect it to share these attributes.

    Attributes:
        router: The custom router, if supplied.
        authenticator: The custom authenticator, if supplied.
    """
    router: typ.Optional[Router] = None
    authenticator: typ.Optional[AuthenticatorType] = None


def load_config(logger: Logger, file: io.TextIOWrapper) -> MailriseConfig:
    """Loads configuration data from a YAML file.

    Args:
        logger: The logger, which will be passed to the `MailriseConfig` instance.
        file: The file handle to load YAML from.

    Returns:
        The `MailriseConfig` instance.
    """
    yml = yaml.load(
        file, Loader=partial(ConfigFileLoader, logger=logger))  # type: ignore
    if not isinstance(yml, dict):
        logger.critical('YAML root node is not a mapping')
        raise SystemExit(1)

    yml_listen = yml.get('listen', {})

    yml_tls = yml.get('tls', {})
    yml_tls_mode = yml_tls.get('mode', 'off').upper()
    try:
        tls_mode = TLSMode[yml_tls_mode]
    except KeyError as exc:
        logger.critical('Invalid TLS operating mode: %s', yml_tls_mode)
        raise SystemExit(1) from exc
    tls_certfile = yml_tls.get('certfile', None)
    tls_keyfile = yml_tls.get('keyfile', None)
    if tls_mode != TLSMode.OFF and not (tls_certfile and tls_keyfile):
        logger.critical('TLS enabled, but certificate and key files not specified')
        raise SystemExit(1)

    yml_smtp = yml.get('smtp', {})

    router = None
    authenticator = None
    yml_import_path = yml.get('import_code', None)
    if yml_import_path:
        logger.info('Importing configurable Python code from: %s', yml_import_path)
        imported = _load_imported_code(logger, yml_import_path)
        if imported.router:
            logger.info('Discovered a custom router')
            router = imported.router
        if imported.authenticator:
            logger.info('Discovered a custom authenticator')
            authenticator = imported.authenticator
    if not router:
        router = load_simple_router(logger, yml.get('configs', {}))
    if not authenticator:
        authenticator = _load_authenticator(yml_smtp.get('auth', {}))

    return MailriseConfig(
        logger=logger,
        listen_host=yml_listen.get('host', ''),
        listen_port=yml_listen.get('port', 8025),
        tls_mode=tls_mode,
        tls_certfile=tls_certfile,
        tls_keyfile=tls_keyfile,
        smtp_hostname=yml_smtp.get('hostname', None),
        router=router,
        authenticator=authenticator
    )


def _load_imported_code(logger: Logger, file_path: str) -> MailriseImportedCode:
    spec = importlib.util.spec_from_file_location(os.path.basename(file_path), file_path)
    if not (spec and spec.loader):
        logger.critical(
            'Nonexistent path or invalid Python when importing code from: %s', file_path)
        raise SystemExit(1)

    module = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(module)
    except Exception as exc:  # pylint: disable=broad-except
        logger.critical('Exception when importing code from: %s', file_path, exc_info=True)
        raise SystemExit(1) from exc

    return typ.cast(MailriseImportedCode, module)


def _load_authenticator(config: dict[str, typ.Any]) -> typ.Optional[AuthenticatorType]:
    if 'basic' in config and isinstance(config['basic'], dict):
        logins = {str(username): str(password)
                  for username, password in config['basic'].items()}
        return typ.cast(AuthenticatorType, BasicAuthenticator(logins=logins))

    return None



"""
A dummy custom router and authenticator for testing and demonstration purposes.
"""

from logging import Logger
from types import SimpleNamespace
from typing import Any, AsyncGenerator

from aiosmtpd.smtp import AuthResult

from mailrise.router import AppriseNotification, EmailMessage, Router


# The typing and inheritance information is not strictly necessary, but it comes
# in handy when developing your code.

class FullyTypedNoopRouter(Router):  # pylint: disable=too-few-public-methods
    """A dummy custom router with full typing information."""
    async def email_to_apprise(
        self, logger: Logger, email: EmailMessage, auth_data: Any, **kwargs) \
            -> AsyncGenerator[AppriseNotification, None]:
        yield AppriseNotification(
            config='urls: ["json://localhost"]',
            title='Hello, World!',
            body='Lorem ipsum dolor sit amet')

class EasyNoopRouter:  # pylint: disable=too-few-public-methods
    """A dummy custom router with the minimum required code."""
    async def email_to_apprise(self, _logger, _email, _auth_data, **_kwargs):
        """Our replacement for email_to_apprise()."""
        # An ordinary dictionary will not work; we need dot notation access.
        yield SimpleNamespace(
            config='urls: ["json://localhost"]',
            title='Hello, World!',
            body='Lorem ipsum dolor sit amet')



# Technically, an authenticator has to satisfy the
# aiosmtpd.smtp.AuthenticatorType type, but there's no way in Python to
# communicate that to mypy.

def noop_authenticator(_server, _session, _envelope, _mechanism, _auth_data):
    """A dummy custom authenticator."""
    return AuthResult(success=False)


router = FullyTypedNoopRouter()
authenticator = noop_authenticator



"""
Tests for the custom Python loader.
"""

import logging
from io import StringIO
from pathlib import Path
from typing import cast

import pytest
from aiosmtpd.smtp import Envelope, Session, SMTP

from mailrise.config import load_config
from mailrise.router import EmailMessage


_logger = logging.getLogger(__name__)
import_path = Path(__file__).parent/'noop_pluggable.py'


@pytest.mark.asyncio
async def test_import_noop() -> None:
    """Tests for the dummy router."""
    file = StringIO(f"""
        import_code: "{import_path}"
    """)
    mrise = load_config(_logger, file)

    router = mrise.router
    notifications = [n async for n
                     in router.email_to_apprise(_logger, cast(EmailMessage, {}), {})]
    assert len(notifications) == 1
    notification = notifications[0]
    assert notification.title == 'Hello, World!'
    assert notification.body == 'Lorem ipsum dolor sit amet'

    authenticator = mrise.authenticator
    assert authenticator is not None
    result = authenticator(cast(SMTP, {}), cast(Session, {}), cast(Envelope, {}), '', {})
    assert not result.success
